Class: SimpleThrottle

Inherits:
Object
  • Object
show all
Defined in:
lib/simple_throttle.rb

Overview

Create a simple throttle that can be used to limit the number of request for a resouce per time period. These objects are thread safe.

Constant Summary collapse

LUA_SCRIPT =

Server side Lua script that maintains the throttle in redis. The throttle is stored as a list of timestamps in milliseconds. When the script is invoked it will scan the oldest entries removing any that should be expired from the list. If the list is below the specified limit then the current entry will be added. The list is marked to expire with the oldest entry so there’s no need to cleanup the lists.

<<~LUA
  local list_key = KEYS[1]
  local limit = tonumber(ARGV[1])
  local ttl = tonumber(ARGV[2])
  local now = ARGV[3]
  local pause_to_recover = tonumber(ARGV[4])
  local amount = tonumber(ARGV[5])
  local cleanup = tonumber(ARGV[6])

  local size = redis.call('llen', list_key)
  if size >= limit or (cleanup > 0 and size > 0) then
    local expired = tonumber(now) - ttl
    while size > 0 do
      local t = redis.call('lpop', list_key)
      if tonumber(t) > expired then
        redis.call('lpush', list_key, t)
        break
      end
      size = size - 1
    end
  end

  if pause_to_recover > 0 then
    limit = limit + 1
  end

  if size + amount > limit then
    amount = (limit - size) + 1
  end

  if size < limit then
    for i = 1, amount do
      redis.call('rpush', list_key, now)
    end
    redis.call('pexpire', list_key, ttl)
  end

  return size + amount
LUA

Instance Attribute Summary collapse

Class Method Summary collapse

Instance Method Summary collapse

Constructor Details

#initialize(name, ttl:, limit:, pause_to_recover: false, redis: nil) ⇒ SimpleThrottle

Create a new throttle.

Parameters:

  • name (String)

    unique name for the throttle

  • ttl (Numeric)

    number of seconds that the throttle will remain active

  • limit (Integer)

    number of allowed requests within the throttle ttl

  • pause_to_recover (Boolean) (defaults to: false)

    require processes calling the throttle to pause at least temporarily before freeing up the throttle. If this is true, then a throttle called constantly with no pauses will never free up.

  • redis (Redis, Proc) (defaults to: nil)

    Redis instance to use or a Proc that yields a Redos instance



142
143
144
145
146
147
148
149
# File 'lib/simple_throttle.rb', line 142

def initialize(name, ttl:, limit:, pause_to_recover: false, redis: nil)
  @name = name.to_s
  @name = name.dup.freeze unless name.frozen?
  @limit = limit.to_i
  @ttl = ttl.to_f
  @pause_to_recover = !!pause_to_recover
  @redis = redis
end

Instance Attribute Details

#limitObject (readonly)

Returns the value of attribute limit.



131
132
133
# File 'lib/simple_throttle.rb', line 131

def limit
  @limit
end

#nameObject (readonly)

Returns the value of attribute name.



131
132
133
# File 'lib/simple_throttle.rb', line 131

def name
  @name
end

#ttlObject (readonly)

Returns the value of attribute ttl.



131
132
133
# File 'lib/simple_throttle.rb', line 131

def ttl
  @ttl
end

Class Method Details

.[](name) ⇒ SimpleThrottle

Returns a globally defined throttle with the specfied name.

Parameters:

  • name (String, Symbol)

    name of the throttle

Returns:



79
80
81
82
83
# File 'lib/simple_throttle.rb', line 79

def [](name)
  if defined?(@throttles) && @throttles
    @throttles[name.to_s]
  end
end

.add(name, ttl:, limit:, pause_to_recover: false, redis: nil) ⇒ void

This method returns an undefined value.

Add a global throttle that can be referenced later with the [] method. This can be used to configure global throttles that you want to setup once and then use in multiple places.

Parameters:

  • name (String)

    unique name for the throttle

  • ttl (Numeric)

    number of seconds that the throttle will remain active

  • limit (Integer)

    number of allowed requests within the throttle ttl

  • pause_to_recover (Boolean) (defaults to: false)

    require processes calling the throttle to pause at least temporarily before freeing up the throttle. If this is true, then a throttle called constantly with no pauses will never free up.

  • redis (Redis, Proc) (defaults to: nil)

    Redis instance to use or a Proc that yields a Redos instance



68
69
70
71
72
73
# File 'lib/simple_throttle.rb', line 68

def add(name, ttl:, limit:, pause_to_recover: false, redis: nil)
  @lock.synchronize do
    @throttles ||= {}
    @throttles[name.to_s] = new(name, limit: limit, ttl: ttl, pause_to_recover: pause_to_recover, redis: redis)
  end
end

.redisRedis

Return the Redis instance where the throttles are stored.

Returns:

  • (Redis)


101
102
103
104
105
106
107
108
# File 'lib/simple_throttle.rb', line 101

def redis
  @redis_client ||= Redis.new
  if @redis_client.is_a?(Proc)
    @redis_client.call
  else
    @redis_client
  end
end

.set_redis(client = nil, &block) ⇒ void

This method returns an undefined value.

Set the Redis instance to use for maintaining the throttle. This can either be set with a hard coded value or by the value yielded by a block. If the block form is used it will be invoked at runtime to get the instance. Use this method if your Redis instance isn’t constant (for example if you’re in a forking environment and re-initialize connections on fork)

Parameters:

  • client (Redis, Proc) (defaults to: nil)

Yield Returns:

  • (Redis)


94
95
96
# File 'lib/simple_throttle.rb', line 94

def set_redis(client = nil, &block)
  @redis_client = (client || block) # rubocop:disable Style/RedundantParentheses
end

Instance Method Details

#allowed!Boolean

Returns true if the limit for the throttle has not been reached yet. This method will also track the throttled resource as having been invoked on each call.

Returns:

  • (Boolean)


155
156
157
158
# File 'lib/simple_throttle.rb', line 155

def allowed!
  size = add_request(1, false)
  size <= limit
end

#increment!(amount = 1) ⇒ Integer

Increment the throttle by the specified and return the current size. Because how the throttle is implemented in Redis, the return value will always max out at the throttle limit + 1 or, if the pause to recover option is set, limit + 2.

Parameters:

  • amount (Integer) (defaults to: 1)

    amount to increment the throttle by

Returns:

  • (Integer)


166
167
168
# File 'lib/simple_throttle.rb', line 166

def increment!(amount = 1)
  add_request(amount, true)
end

#peekInteger

Peek at the current number for throttled calls being tracked.

Returns:

  • (Integer)


180
181
182
183
184
# File 'lib/simple_throttle.rb', line 180

def peek
  timestamps = redis_client.lrange(redis_key, 0, -1).collect(&:to_i)
  min_timestamp = ((Time.now.to_f - ttl) * 1000).ceil
  timestamps.count { |t| t > min_timestamp }
end

#reset!void

This method returns an undefined value.

Reset a throttle back to zero.



173
174
175
# File 'lib/simple_throttle.rb', line 173

def reset!
  redis_client.del(redis_key)
end

#wait_timeFloat

Returns when the next resource call should be allowed. Note that this doesn’t guarantee that calling allow! will return true if the wait time is zero since other processes or threads can claim the resource.

Returns:

  • (Float)


191
192
193
194
195
196
197
198
199
200
# File 'lib/simple_throttle.rb', line 191

def wait_time
  if peek < limit
    0.0
  else
    first = redis_client.lindex(redis_key, 0).to_f / 1000.0
    delta = Time.now.to_f - first
    delta = 0.0 if delta < 0
    ttl - delta
  end
end